
#include"table.h"

union table_node {
	struct table table;
	union table_node *next;
};

struct table_pool {
	union table_node *base, *avail;
	lock_t free;
};

static struct table_pool table_pool; /* This is common to all threads */
volatile struct table_op *table_op[TABLE_TYPE_MAX]; /* This is common to all process's */

table_t table_new(int type, int size, int width, compare_fn compare, hash_fn hash_rehash, float occupancy) {
	union table_node *new;
	static void *base;
	int cancel;

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
	if(lock(&table_pool->free)) {
		if(table_pool->avail) {
			new = table_pool->avail;
			table_pool->avail = table_pool->avail.next;
			if(base = malloc(width * size)) {
				new->base.data = new->base.avail = base;
				unlock(&table_pool->free);
				if(new->visit = visit_new()) {
					new->table.type = type;
					new->table.size[0] = new->table.size[1] = size;
					new->size[3] = width;
					if(table_op[(type / 16) % TABLE_TYPE_MAX]->new(new)) {
						void **element;
						int pos;

						element = new->base.data;
						for(pos = 0; pos < size; pos++) {
							*element = element + width;
							element = *element;
						}
						new->count[0] = new->count[1] = 0;
						new->compare = compare;
						if(type) {
							new->parameter.hash.occupancy =
								occupancy;
							new->parameter.hash.fn.hash =
								hash_rehash;
						}
						new->inuse = TRUE;
						lock_init(&new->lock);
						counter_init(&new->counter, BIASED_PARALLEL);
						pthread_setcancelstate(cancel, NULL);
						return (table_t) &new->table;
					}
					visit_dispose(new->visit);
				}
				free(new->base.data);
			}
			lock(&table_pool->free);
			new->next = table_pool->avail;
			table_pool->avail = new;
		}
		unlock(&table_pool->free);
	}
	pthread_setcancelstate(cancel, NULL);
	return NULL;
}

void table_dispose(table_t table) {
	union table_node *old;
	int type;

	if(table->inuse) {
		counter_up(&table->counter, SEQUENTIAL);
		visit_dispose(table->visit);
		table->inuse = FALSE;
		type = table->type;
		table_op[(type / 32) % TABLE_TYPE_MAX]->dispose(table);
		lock(&table_pool->free);
		old = (union table_node *)table;
		old->next = table_pool.avail;
		table_pool.avail = old;
		unlock(&table_pool.free);
		counter_down(&table->counter, SEQUENTIAL);
	}
	return;
}

void *table_lookup(table_t table, void *element, tableptr_t *ptr, int lookup) {
	if(table->inuse) {
		if(lookup == LOOKUP_DELETE && ! table_count(table)) errno = EEMPTY;
		else {
			if(lookup == LOOKUP_INSERT && table_count(table) == table_size(table)) {
				errno = EFULL;
			}
			else {
				void *found;
				int type, cancel;

				if(lookup == LOOKUP_INSERT) {
					lock(&table->free);
					item = *table->base.avail;
					table->base.avail = *table->base.avail;
					unlock(&table->free);
					memcpy(item, element, table->size[2]);
					item -= table->base.data;
					item /= table->size[2];
				}
				else item = element;
				pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
				counter_up(&table->counter, PARALLEL);
				type = table->type;
				found = table_op[(type / 32) % TABLE_TYPE_MAX]->lookup(table, item, ptr, lookup);
				if(! found) {
					errno = ENOTFOUND;
					if(lookup == LOOKUP_INSERT) {
						item = table->base.data + (item * table->size[2]);
						lock(&table->free);
						*(void **)item = (void *)table->base.avail;
						table->base.avail = (void **)item;
						unlock(&table->free);
						if((table->type / NO_DUPLICATES) % 2)
							errno = EFOUND;
					}
				}
				else {
					found = table->base.data + (found * table->size[2]);
					if(lookup == LOOKUP_INSERT) table->count[0]++;
					else {
						if(lookup == LOOKUP_DELETE) table->count[1]++;
						else {
							if(lookup == LOOKUP_REPLACE) {
								lock(&table->free);
								*(void **)found =
									(void *)table->base.avail;
								table->base.avail =
									(void **)found;
								unlock(&table->free);
							}
						}
					}
					memcpy(element, found, table->size[2]);
				}
				counter_down(&table->counter, PARALLEL);
				pthread_setcancelstate(cancel, NULL);
				return element;
			}
		}
	}
	return NULL;
}

void *table_get(table_t table, void *element, tableptr_t *ptr, int get) {
	void *found;
	int type;

	if(table->inuse) {
		counter_up(&table->counter, PARALLEL);
		type = table->type;
		if(found = table_op[(type / 32) % TABLE_TYPE_MAX]->get(table, ptr, get)) {
			memcpy(element, found, table->size[2]);
			counter_down(&table->counter, PARALLEL);
			return element;
		}
		counter_down(&table->counter, PARALLEL);
	}
	return NULL;
}

table_t table_resize(table_t table, int size) {
	static void *base;
	int retries, success, cancel, type;

	if(table->inuse) {
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
		changed = NULL;
		retries = AUTO_RESIZE_RETRIES;
		while(retries-- >= 0) {
			if(success = counter_tryup(&table->counter, SEQUENTIAL)) {
				if(success == 2)
					while(counter->biased.parallel) pthread_switch();
				type = table->type;
				if(base = realloc(table->size[2] * size)) {
					if(table_op[(type / 32) % TABLE_TYPE_MAX]->resize(table, size)) {
						table->base.avail -= table->base.data;
						table->base.avail /= table->size[2];
						table->base.data = base;
						table->base.avail = table->base.data + (table->base.avail * table->size[2]);
						table->size[0] = size;
						counter_down(&table->counter, SEQUENTIAL);
						pthread_setcancelstate(cancel, NULL);
						return table;
					}
					base = realloc(table->size[2] * table->size[1]);
				}
			}
			counter_down(&table->counter, SEQUENTIAL);
			break;
		}
		pthread_switch();
	}
	pthread_setcancelstate(cancel, NULL);
	return NULL;
}
