#ifndef LIST_H
#define LIST_H

#ifndef LIST_MAX
#define LIST_MAX (100)
#endif

enum list_insert_type { NO_INSERT_FIRST = 1, NO_INSERT_MIDDLE = 2, NO_INSERT_LAST = 4 };
enum list_delete_type { NO_DELETE_FIRST = 8, NO_DELETE_MIDDLE = 16, NO_DELETE_LAST = 32 };
enum linked_type { DOUBLY_LINKED, SINGLY_LINKED = 64, NON_CIRCULAR, CIRCULAR = 128 };

#define STACK_TYPE (54)
#define QUEUE_TYPE (51)
#define LINKED_TYPE (0)

#define list_array(type) (((type) & NO_INSERT_MIDDLE) && ((type) &  NO_DELETE_MIDDLE))

struct single_node {
	void *element;
	struct single_node *next;
};

struct double_node {
	void *element;
	struct double_node *previous, *next;
};

union linked_node {
	struct single_node *single;
	struct double_node *double_;
};

struct linked_base {
	union linked_node base, avail;
};

union list_base {
	void *array;
	struct linked_base linked;
};

union list_end {
	struct array_end {
		void *first, *last;
	} array;
	struct linked_end {
		union linked_node first, last;
	} linked;
};

struct list {
	int type, size, len;
	union list_base base;
	union list_end end;
};

union list_node {
	struct list list;
	union list_node *next;
};

struct list_pool {
	union list_node base[LIST_MAX], *avail;
	visit_t visit;
	lock_t free;
};

static struct list_pool list_pool;

typedef const struct list * const list_t;

struct linked_ptr {
	union linked_node current, previous;
};

typedef union list_ptr {
	void *array;
	struct linked_ptr linked;
} listptr_t;

static struct list_pool list_pool;

#define insdel(type, insdel) (((type) / (insdel)) % 2)
#define array(type) (no_insdel(type, NO_INSERT_MIDDLE) && no_insdel(type, NO_DELETE_MIDDLE))
#define single(type) (((type) / SINGLY_LINKED) % 2)

listptr_t *listptr_init(listptr_t *ptr, int type) {
#assert(ptr != NULL);
	if(array(type)) ptr->array = NULL;
	else {
		ptr->linked.current = NULL;
		if(single(type)) ptr->linked.previous = NULL;
	}
	return ptr;
}

list_t list_new(int type, int size) {
	static union linked_node linked;
	static void *array;
	bool_t isarray;
	list_t new;

#assert(size > 0);
	if(lock(&list_pool->free) > 0) {
		if(new = list_pool->avail) {
			isarray = array(type);
			single = single(type);
			if((isarray && base = malloc(PTR_SIZE * size)) ||
			(base = malloc((single ? SINGLE_SIZE : DOUBLE_SIZE) * size))) {
				unlock(&list_pool->free);
				if(isarray) {
					new->base.array = base;
					new->end.array.first =
					new->end.array.last = NULL;
				}
				else {
					if(single) {
						linked.single =
						new->base.linked.base.single =
						new->base.linked.avail.single = base;
					}
					else {
						linked.double =
						new->base.linked.base.double =
						new->base.linked.avail.double = base;
					}
					for(pos = 0; pos < size - 1; pos++) {
						if(single) {
							linked.single->next =
							linked.single + 1;
							++linked.single;
						}
						else {
							linked.double->previous =
							linked.double - 1;
							linked.double->next =
							linked.double + 1;
							++linked.double;
						}
					}
				}
				lock_init(&new->free);
				counter_init(&new->counter, BIASED_PARALLEL);
				unlock(&list_pool->free);
				return new;
			}
			new->next = list_pool->avail;
			list_pool->avail = new;
		}
		unlock(&list_pool->free);
	}
	return NULL;
}

void list_dispose(list_t list) {
#assert(list != NULL);
	if(list->inuse) {
		counter_up(&list->counter, SEQUENTIAL);
		list->inuse = FALSE;
		if(lock(&list_pool->free) > 0) {
			if(array(list->type)) {
				free(list->base.array);
			}
			else {
				if(single(list->type)) {
					free(list->base.linked.single);
				}
				else {
					free(list->base.linked._double);
				}
			}
			(union list_node *)list->next = list_pool->avail;
			list_pool->avail = list;
			counter_down(&list->counter, SEQUENTIAL);
			unlock(&list_pool->free);
		}
	}
	return;
}

#define list_bound(list, ptr) (array(list->type) ?
	bound_check(ptr->array, list->base.array, list->size) : single(list->type) ?
	bound_check(ptr->linked.current.single, list->base.linked.single, list->size) :
	bound_check(ptr->linked.current._double, list->base.linked._double, list->size))

list_t list_insert(list_t list, void *element, listptr_t *ptr) {
#assert(list != NULL && element != NULL && list_bound(list, ptr));
	if(list->inuse) {
		isarray = array(list->type);
		single = single(list->type);
		if(! ptr || (isarray && ptr->array == list->end.array.first) ||
			(! isarray && single ?
			 ptr->linked.current.single == list->end.linked.first.single :
			 ptr->linked.current._double == list->end.linked.first._double)) {
			if(insdel(list->type, NO_INSERT_FIRST)) return NULL;
		}
		else {
			if(ptr) {
				if((isarray && ptr->array == list->end.array.last) ||
					(! isarray && single ? ptr->linked.current.single ==
					 list->end.linked.last.single :
					 ptr->linked.current._double ==
					 list->end.linked.last._double)) {
					list_append(list, element, ptr);
				}
				else {
					if(insdel(list->type, NO_INSERT_MIDDLE)) return NULL;
				}
			}
			else return NULL;
		}
		if(ptr) {
			if(single) {
				if(! ptr->linked.previous.single ||
					ptr->linked.previous.single->next !=
					ptr->linked.current.single) {
					struct single *current;

					current = ptr->linked.current.single;
					if(list_get(list, GET_FIRST, ptr) != current) {
						while(list_get(list, GET_NEXT, ptr) != current);
					}
				}
				if(ptr->linked.previous.single) {
					ptr->linked.previous.single->next = new;
				}
				else {
					list->end.linked.first.single = new;
				}
				new->next = ptr->linked.current.single;
			}
			else {
			}
		}
	}
}

list_t list_append(list_t list, void *element, listptr_t *ptr) {
}

list_t list_delete(list_t list, listptr_t *ptr) {
}

list_t list_execute(list_t list, listptr_t *ptr) {
}

void *list_get(list_t list, int get, listptr_t *ptr) {
}

list_t list_resize(list_t list, int size) {
}

#define list_type(list) ((list)->type)
#define list_size(list) ((list)->size)
#define list_count(list) ((list)->count)

#endif
