
#include<pthread.h>

#include"counter.h"

struct counter_entry {
	char op;
	counter_t *counter;
};

union counter_node {
	struct counter_entry entry;
	union counter_node *next;
};

#define COUNTER_NODE_MAX (PTHREAD_MAX * 2)

struct counter_pool {
	union counter_node base[COUNTER_NODE_MAX], *avail;
	lock_t free;
};

static struct counter_pool counter_pool;

void counter_init(counter_t *counter, int type) {
	pthread_atomic_t atomic;
	pthread_cancel_t cancel;

	pthread_setatomicstate(PTHREAD_ATOMIC, &atomic);
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
	counter->rwlock = 0;
	counter->type = type;
	if(type == BALANCED) {
		counter->balanced.sequential.up = counter->balanced.sequential.down = 0;
		counter->balanced.parallel.up = counter->balanced.parallel.down = 0;
	}
	else {
		counter->biased.sequential = counter->biased.parallel = 0;
	}
	pthread_setcancelstate(cancel, NULL);
	pthread_setatomicstate(atomic, NULL);
	return;
}

static struct counter_entry *find_counter(counter_t *counter, int op, listptr_t *ptr) {
	struct counter_entry *request;
	pthread_t self;

	self = pthread_self();
	request = list_get(self->list.counter, GET_FIRST, ptr);
	while(request && (request->counter != counter || request->op < op)) {
		request = list_get(self->list.counter, GET_NEXT, ptr);
	}
	return request;
}

static int add_counter(counter_t *counter, int op) {
	union counter_node *avail;
	pthread_t self;

	self = pthread_self();
	lock(&counter_pool.free);
	if(avail = counter_pool.avail) {
		counter_pool.avail = avail->next;
		unlock(&counter_pool.free);
		avail->counter.op = sequential;
		avail->counter.counter = counter;
		if(list_insert(self->list.counter, &avail->counter, NULL)) return 1;
		lock(&counter_pool.free);
		avail->next = counter_pool.avail;
		counter_pool.avail = avail;
	}
	unlock(&counter_pool.free);
	return 0;
}

int counter_tryup(counter_t *counter, int op) {
	int parallel_down, sequential_down, success;
	pthread_cancel_t cancel;
	pthread_t self;

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
	self = pthread_self();
	if(op == PARALLEL) {
		if(counter->type == BALANCED) {
			sequential_down = counter->balanced.sequential.up;
			counter->balanced.parallel.up++;
			success = counter->balanced.sequential.down >= sequential_down ? 1 : 0;
		}
		else {
			counter->biased.parallel++;
			if(counter->type == BIASED_PARALLEL) {
				success = (counter->biased.sequential != 1 ||
					counter->biased.parallel != 1) ? 1 : 0;
			}
			else {
				success = (counter->biased.sequential != 1) ? 1 : 0;
			}
		}
	}
	else {
		if(counter->type == BALANCED) {
			sequential_down = counter->balanced.sequential.up;
			counter->balanced.sequential.up++;
			parallel_down = counter->balanced.parallel.up;
			success = (counter->balanced.parallel.down >= parallel_down &&
				counter->balanced.sequential.down >= sequential_down) ? 1 : 0;
		}
		else {
			if(! counter->biased.sequential) {
				counter->biased.sequential++;
				if(success = (counter->biased.sequential == 1))
					success = (counter->biased.parallel) ? 2 : 1;
			}
		}
	}
	if(success && ! counter->rwlock) {
		if(! entry = find_counter(counter, NULL)) {
			if(! add_counter(counter, op)) success = -1;
		}
		else success = -2;
	}
	if(success <= 0) {
		if(op == PARALLEL) {
			if(counter->type == BALANCED) counter->balanced.parallel.up--;
			else counter->biased.parallel--;
		}
		else {
			if(counter->type == BALANCED) counter->balanced.sequential.up--;
			else counter->biased.sequential--;
		}
		if((success == -2) && (op == entry->op)) success = 1;
	}
	pthread_setcancelstate(cancel, NULL);
	return success;
}

int counter_timedup(counter_t *counter, int op, struct timespec *abstime) {
	pthread_cancel_t cancel;
	int success;

	while(((success = counter_tryup(counter, op)) <= 0) && ! (abstime && past(abstime))) {
		pthread_switch();
	}
	if(success == 2) {
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
		while(counter->biased.parallel) {
			if(abstime && past(abstime)) return counter_down(counter, op);
			pthread_switch();
		}
		pthread_setcancelstate(cancel, NULL);
	}
	return 1;
}

int counter_down(counter_t *counter, int op) {
	struct counter_entry *entry;
	union counter_node *avail;
	pthread_cancel_t cancel;
	pthread_t self;
	listptr_t ptr;

	if(! counter->rwlock) {
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
		self = pthread_self();
		while((entry = find_counter(counter, op, &ptr)) &&
			(! list_delete(self->list.counter, &ptr)));
		if(entry) {
			avail = (union counter_node *)entry;
			lock(&rwlock_pool.free);
			avail->next = rwlock_pool.avail;
			rwlock_pool.avail = avail;
			unlock(&rwlock_pool.free);
		}
		else {
			pthread_setcancelstate(cancel, NULL);
			return 0;
		}
	}
	(counter->type == BALANCED) ?  ((op == PARALLEL) ?
	++counter->balanced.parallel.down : ++counter->balanced.sequential.down) :
	(op == PARALLEL) ? --counter->biased.parallel : --counter->biased.sequential;
	pthread_setcancelstate(cancel, NULL);
	return 1;
}

int counter_active(counter_t *counter) {
	pthread_atomic_t atomic;
	int active;

	pthread_setatomicstate(PTHREAD_ATOMIC, &atomic);
	active = counter->type == BALANCED ?
		counter->balanced.parallel.up != counter->balanced.parallel.down ||
		counter->balanced.sequential.up != counter->balanced.sequential.down :
		counter->biased.parallel || counter->biased.sequential;
	pthread_setatomicstate(atomic, NULL);
	return active;
}
